#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
import os
import tempfile
import time
from urllib.parse import urlparse
import yaml
import zipfile

from novaclient import client as nova_client
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from tacker.objects import fields
from tacker.tests.functional import base
from tacker.tests.functional.common.fake_server import FakeServerManager
from tacker.tests import utils
from tacker.vnfm.infra_drivers.openstack import constants as infra_cnst


VNF_PACKAGE_UPLOAD_TIMEOUT = 60
VNF_INSTANTIATE_TIMEOUT = 60
VNF_TERMINATE_TIMEOUT = 60
VNF_SUBSCRIPTION_TIMEOUT = 60
VNF_INSTANTIATE_ERROR_WAIT = 80
VNF_DELETE_COMPLETION_WAIT = 60
VNF_HEAL_TIMEOUT = 600
VNF_LCM_DONE_TIMEOUT = 1200
RETRY_WAIT_TIME = 5
FAKE_SERVER_MANAGER = FakeServerManager()
FAKE_SERVER_PORT = 9990
MOCK_NOTIFY_CALLBACK_URL = '/notification/callback'
UUID_RE = r'\w{8}-\w{4}-\w{4}-\w{4}-\w{12}'


def _get_external_virtual_links(net0_id):
    return [
        {
            "id": "net0",
            "resourceId": net0_id,
            "extCps": [{
                "cpdId": "CP1",
                "cpConfig": [{
                    "cpProtocolData": [{
                        "layerProtocol": "IP_OVER_ETHERNET",
                    }]
                }]
            }]
        }
    ]


def _create_csar_user_data_common(csar_dir):
    ud_common_dir = os.path.join(csar_dir, "../user_data_common/")
    return _create_csar_with_unique_vnfd_id(
        csar_dir, ud_common_dir)


def _create_csar_with_unique_vnfd_id(csar_dir, *include_dirs):
    tempfd, tempname = tempfile.mkstemp(suffix=".zip",
        dir=os.path.dirname(csar_dir))
    os.close(tempfd)

    common_dir = os.path.join(csar_dir, "../common/")
    target_dirs = [csar_dir, common_dir]
    target_dirs.extend(include_dirs)

    unique_id = uuidutils.generate_uuid()
    with zipfile.ZipFile(tempname, 'w') as zcsar:
        _write_zipfile(zcsar, unique_id, target_dirs)

    return tempname, unique_id


def _write_zipfile(zcsar, unique_id, target_dir_list):
    for target_dir in target_dir_list:
        for (dpath, _, fnames) in os.walk(target_dir):
            if not fnames:
                continue
            for fname in fnames:
                src_file = os.path.join(dpath, fname)
                dst_file = os.path.relpath(
                    os.path.join(dpath, fname), target_dir)
                if fname.endswith('.yaml') or fname.endswith('.yml'):
                    with open(src_file, 'rb') as yfile:
                        data = yaml.safe_load(yfile)
                        utils._update_unique_id_in_yaml(data, unique_id)
                        zcsar.writestr(dst_file, yaml.dump(
                            data, default_flow_style=False,
                            allow_unicode=True))
                else:
                    zcsar.write(src_file, dst_file)


def _create_and_upload_vnf_package(
        tacker_client,
        user_defined_data,
        temp_csar_path):
    # create vnf package
    body = jsonutils.dumps({"userDefinedData": user_defined_data})
    resp, vnf_package = tacker_client.do_request(
        '/vnfpkgm/v1/vnf_packages', "POST", body=body)

    with open(temp_csar_path, 'rb') as file_object:
        resp, resp_body = tacker_client.do_request(
            '/vnfpkgm/v1/vnf_packages/{id}/package_content'.format(
                id=vnf_package['id']),
            "PUT", body=file_object, content_type='application/zip')

    # wait for onboard
    timeout = VNF_PACKAGE_UPLOAD_TIMEOUT
    start_time = int(time.time())
    show_url = os.path.join('/vnfpkgm/v1/vnf_packages', vnf_package['id'])
    vnfd_id = None
    while True:
        resp, body = tacker_client.do_request(show_url, "GET")
        if (200 <= resp.status_code < 300) and (
                body['onboardingState'] == "ONBOARDED"):
            vnfd_id = body['vnfdId']
            break

        if ((int(time.time()) - start_time) > timeout):
            raise TimeoutError("Failed to onboard vnf package")

        time.sleep(1)

    # remove temporarily created CSAR file
    os.remove(temp_csar_path)
    return vnf_package['id'], vnfd_id


def _delete_vnf_package(tacker_client, vnf_package_id):
    url = '/vnfpkgm/v1/vnf_packages/%s' % vnf_package_id

    # Update vnf package before delete
    req_body = jsonutils.dumps({"operationalState": "DISABLED"})
    tacker_client.do_request(url, "PATCH", body=req_body)

    # Delete vnf package before delete
    tacker_client.do_request(url, "DELETE")


def _show_vnf_package(tacker_client, vnf_package_id):
    # wait for onboard
    timeout = VNF_PACKAGE_UPLOAD_TIMEOUT
    start_time = int(time.time())
    show_url = os.path.join('/vnfpkgm/v1/vnf_packages', vnf_package_id)
    while True:
        resp, body = tacker_client.do_request(show_url, "GET")
        if resp.ok:
            return resp, body

        if ((int(time.time()) - start_time) > timeout):
            raise TimeoutError("Failed to onboard vnf package")

        time.sleep(1)


def _list_vnf_package(tacker_client, **kwargs):
    # wait for onboard
    timeout = VNF_PACKAGE_UPLOAD_TIMEOUT
    start_time = int(time.time())
    while True:
        resp, body = tacker_client.do_request(
            '/vnfpkgm/v1/vnf_packages', "GET", **kwargs)
        if resp.ok:
            return resp, body

        if ((int(time.time()) - start_time) > timeout):
            raise TimeoutError("Failed to onboard vnf package")

        time.sleep(1)


def _create_instantiate_vnf_request_body(flavour_id,
        instantiation_level_id=None, vim_id=None, ext_vl=None,
        add_params=None):
    request_body = {"flavourId": flavour_id}

    if instantiation_level_id:
        request_body["instantiationLevelId"] = instantiation_level_id

    if ext_vl:
        request_body["extVirtualLinks"] = ext_vl

    if vim_id:
        request_body["vimConnectionInfo"] = [
            {"id": uuidutils.generate_uuid(),
                "vimId": vim_id,
                "vimType": "ETSINFV.OPENSTACK_KEYSTONE.v_2"}]

    if add_params:
        request_body["additionalParams"] = add_params

    return request_body


class BaseVnfLcmTest(base.BaseTackerTest):

    is_setup_error = False
    # NOTE: If prepare_fake_server is set(by default) then this base
    # class will prepare(create and start) the fake http server which
    # takes time and can add up more time in gate CI. Any child class
    # can set it false. For example, BaseVnfLcmMultiTenantTest which
    # create their own servers for two different tenants.
    prepare_fake_server = True

    @classmethod
    def setUpClass(cls):
        '''Set up test class.

        we set up fake NFVO server for test at here.
        '''
        super(BaseVnfLcmTest, cls).setUpClass()
        if cls.prepare_fake_server:
            cls._prepare_start_fake_server(FAKE_SERVER_MANAGER,
                FAKE_SERVER_PORT)

    @classmethod
    def tearDownClass(cls):
        super(BaseVnfLcmTest, cls).tearDownClass()
        if cls.prepare_fake_server:
            FAKE_SERVER_MANAGER.stop_server()

    def setUp(self):
        super(BaseVnfLcmTest, self).setUp()

        if self.is_setup_error:
            self.fail("Faild, not exists pre-registered image.")

        if self.prepare_fake_server:
            callback_url = os.path.join(MOCK_NOTIFY_CALLBACK_URL,
                self._testMethodName)
            self._clear_history_and_set_callback(FAKE_SERVER_MANAGER,
                callback_url)

        self.tacker_client = base.BaseTackerTest.tacker_http_client()

        self.base_vnf_instances_url = "/vnflcm/v1/vnf_instances"
        self.base_subscriptions_url = "/vnflcm/v1/subscriptions"
        self.base_vnf_lcm_op_occs_url = "/vnflcm/v1/vnf_lcm_op_occs"

        vim_list = self.client.list_vims()
        self.vim = self.get_vim(vim_list, 'VIM0')
        if not self.vim:
            assert False, "vim_list is Empty: Default VIM is missing"
        result = self._create_network_settings()
        self.ext_networks = result.get('ext_networks')
        self.ext_vl = result.get('ext_vl')
        self.ext_mngd_networks = result.get('ext_mngd_networks')
        self.ext_link_ports = result.get('ext_link_ports')
        self.ext_subnets = result.get('ext_subnets')
        self.changed_ext_networks = result.get('changed_ext_networks')
        self.changed_ext_subnets = result.get('changed_ext_subnets')

    @classmethod
    def _prepare_start_fake_server(cls, fake_server,
            fake_server_port):
        fake_server.prepare_http_server(address='localhost',
                port=fake_server_port)
        fake_server.start_server()

    def _clear_history_and_set_callback(self,
            fake_server_manager,
            callback_url):
        fake_server_manager.clear_history(callback_url)
        fake_server_manager.set_callback(
            'POST',
            callback_url,
            status_code=204
        )
        fake_server_manager.set_callback(
            'GET',
            callback_url,
            status_code=204
        )

    def _create_network_settings(self, neutron_client=None):
        if neutron_client is None:
            neutron_client = self.neutronclient()
        # Create external external.
        ext_networks = list()
        # Create external managed networks
        ext_mngd_networks = list()  # Store ids for cleaning.
        # Create external link ports in net0
        ext_link_ports = list()
        # Create external subnet in net1
        ext_subnets = list()  # Store ids for cleaning.
        # Create external networks to change.
        changed_ext_networks = list()
        changed_ext_subnets = list()  # Store ids for cleaning.

        networks = neutron_client.list_networks()
        for nw in networks.get('networks'):
            if nw['name'] == 'net0':
                ext_networks.append(nw['id'])
                ext_vl = _get_external_virtual_links(nw['id'])
                ext_subnets.append(nw['subnets'][0])
                ext_link_ports.append(self._create_port(nw['id'],
                                                 neutron_client))
                ext_link_ports.append(self._create_port(nw['id'],
                                                 neutron_client))
            elif nw['name'] == 'net1':
                ext_mngd_networks.append(nw['id'])

        # create new network.
        ext_net_id, ext_net_name = \
            self._create_network("external_net", neutron_client)
        ext_networks.append(ext_net_id)
        ext_mngd_net_id, _ = \
            self._create_network("external_managed_internal_net",
                    neutron_client)
        ext_mngd_networks.append(ext_mngd_net_id)
        changed_ext_net_id, changed_ext_net_name = \
            self._create_network("changed_external_net", neutron_client)
        changed_ext_networks.append(changed_ext_net_id)

        # Chack how many networks are created.
        networks = neutron_client.list_networks()
        for nw in networks.get('networks'):
            if nw['name'] not in [ext_net_name, changed_ext_net_name]:
                continue

            elif nw['name'] == ext_net_name:
                ext_subnets.append(
                    self._create_subnet(nw,
                                        cidr="22.22.1.0/24",
                                        gateway="22.22.1.1",
                                        neutron_client=neutron_client))
            elif nw['name'] == changed_ext_net_name:
                changed_ext_subnets.append(
                    self._create_subnet(nw,
                                        cidr="22.22.2.0/24",
                                        gateway="22.22.2.1",
                                        neutron_client=neutron_client))
        return {'ext_networks': ext_networks,
                'ext_vl': ext_vl,
                'ext_mngd_networks': ext_mngd_networks,
                'ext_link_ports': ext_link_ports,
                'ext_subnets': ext_subnets,
                'changed_ext_networks': changed_ext_networks,
                'changed_ext_subnets': changed_ext_subnets}

    @classmethod
    def _list_glance_image(cls, filter_name='cirros-0.5.2-x86_64-disk',
            glance_client=None):
        if glance_client is None:
            glance_client = cls.glance_client
        try:
            images = glance_client.images.list()
        except Exception:
            print("glance-image does not exists.", flush=True)
            return []

        if filter_name is None:
            return images

        return list(filter(lambda image: image.name == filter_name, images))

    @classmethod
    def _get_glance_image(cls, image_id, glance_client=None):
        if glance_client is None:
            glance_client = cls.glance_client
        try:
            image = glance_client.images.get(image_id)
        except Exception:
            print("glance-image does not exists.", image_id, flush=True)
            return None

        return image

    @classmethod
    def _create_glance_image(cls, image_data, file_url, glance_client=None):
        if glance_client is None:
            glance_client = cls.glance_client
        image = glance_client.images.create(**image_data)
        glance_client.images.upload(image.id, file_url)

        return image.id

    def _get_glance_image_list_from_stack_resource(
            self, stack_id, stack_resource_name, h_client=None):
        if h_client is None:
            h_client = self.h_client
        image_id_list = []
        for resource_name in stack_resource_name:
            resource_details = self._get_heat_resource(stack_id,
                    resource_name,
                    h_client)
            image = self._get_image_id_from_resource_attributes(
                resource_details)
            if image:
                image_id_list.append(image.id)

        return image_id_list

    @classmethod
    def _list_zone(cls):
        try:
            zone = cls.nova_client.services.list()
        except nova_client.exceptions.ClientException:
            print("availability zone does not exists.", flush=True)
            return []
        return zone

    def _register_subscription(self, request_body, http_client=None):
        if http_client is None:
            http_client = self.http_client
        resp, response_body = http_client.do_request(
            self.base_subscriptions_url,
            "POST",
            body=jsonutils.dumps(request_body))
        return resp, response_body

    def _delete_subscription(self, subscription_id, tacker_client=None):
        if tacker_client is None:
            tacker_client = self.tacker_client
        delete_url = os.path.join(self.base_subscriptions_url, subscription_id)
        resp, body = tacker_client.do_request(delete_url, "DELETE")

        return resp, body

    def _show_subscription(self, subscription_id, tacker_client=None):
        if tacker_client is None:
            tacker_client = self.tacker_client
        show_url = os.path.join(self.base_subscriptions_url, subscription_id)
        resp, body = tacker_client.do_request(show_url, "GET")

        return resp, body

    def _list_subscription(self, tacker_client=None):
        if tacker_client is None:
            tacker_client = self.tacker_client
        resp, body = tacker_client.do_request(
            self.base_subscriptions_url, "GET")

        return resp, body

    def _list_subscription_filter(self, http_client=None, **kwargs):
        if http_client is None:
            http_client = self.http_client
        params = kwargs.get('params', {})
        filter_variable = params['filter']
        subscriptions_list_filter_url = "%s?%s" % (
            self.base_subscriptions_url, filter_variable)

        resp, subscription_body = http_client.do_request(
            subscriptions_list_filter_url, "GET")

        return resp, subscription_body

    def _create_vnf_instance(self, vnfd_id, vnf_instance_name=None,
            vnf_instance_description=None, http_client=None):
        if http_client is None:
            http_client = self.http_client
        request_body = {'vnfdId': vnfd_id}
        if vnf_instance_name:
            request_body['vnfInstanceName'] = vnf_instance_name

        if vnf_instance_description:
            request_body['vnfInstanceDescription'] = vnf_instance_description

        return self._create_vnf_instance_from_body(request_body, http_client)

    def _create_vnf_instance_from_body(self, request_body, http_client=None):
        if http_client is None:
            http_client = self.http_client
        request_body['vnfInstanceName'] = self._testMethodName
        resp, response_body = http_client.do_request(
            self.base_vnf_instances_url,
            "POST",
            body=jsonutils.dumps(request_body))

        return resp, response_body

    def _delete_vnf_instance(self, id, http_client=None):
        if http_client is None:
            http_client = self.http_client
        url = os.path.join(self.base_vnf_instances_url, id)
        resp, body = http_client.do_request(url, "DELETE")

        return resp, body

    def _show_vnf_instance(self, id, http_client=None):
        if http_client is None:
            http_client = self.http_client
        show_url = os.path.join(self.base_vnf_instances_url, id)
        resp, vnf_instance = http_client.do_request(show_url, "GET")

        return resp, vnf_instance

    def _list_vnf_instance(self, http_client=None, **kwargs):
        if http_client is None:
            http_client = self.http_client
        resp, vnf_instances = http_client.do_request(
            self.base_vnf_instances_url, "GET", **kwargs)

        return resp, vnf_instances

    def _wait_vnf_instance(self, id,
            http_client=None,
            instantiation_state=fields.VnfInstanceState.INSTANTIATED,
            timeout=VNF_INSTANTIATE_TIMEOUT):
        if http_client is None:
            http_client = self.http_client
        start_time = int(time.time())
        while True:
            resp, body = self._show_vnf_instance(id, http_client)
            if body['instantiationState'] == instantiation_state:
                break

            if ((int(time.time()) - start_time) > timeout):
                error = ("Vnf instance %(id)s status is %(current)s, "
                         "expected status should be %(expected)s")
                self.fail(error % {"id": id,
                    "current": body['instantiationState'],
                    "expected": instantiation_state})

            time.sleep(5)

    def _instantiate_vnf_instance(self, id, request_body, http_client=None):
        if http_client is None:
            http_client = self.http_client
        url = os.path.join(self.base_vnf_instances_url, id, "instantiate")
        resp, body = http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))

        return resp, body

    def _heal_vnf_instance(self, vnf_instance_id, request_body,
            http_client=None):
        if http_client is None:
            http_client = self.http_client
        url = os.path.join(
            self.base_vnf_instances_url,
            vnf_instance_id,
            "heal")
        resp, body = http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))

        return resp, body

    def _scale_vnf_instance(self, vnf_instance_id, request_body,
            http_client=None):
        if http_client is None:
            http_client = self.http_client
        url = os.path.join(
            self.base_vnf_instances_url,
            vnf_instance_id,
            "scale")
        resp, body = http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))

        return resp, body

    def _terminate_vnf_instance(self, id, request_body, http_client=None):
        if http_client is None:
            http_client = self.http_client
        url = os.path.join(self.base_vnf_instances_url, id, "terminate")
        resp, body = http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))

        return resp, body

    def _update_vnf_instance(self, vnf_instance_id, request_body,
            http_client=None):
        if http_client is None:
            http_client = self.http_client
        url = os.path.join(self.base_vnf_instances_url, vnf_instance_id)
        resp, body = http_client.do_request(url, "PATCH",
                body=jsonutils.dumps(request_body))

        return resp, body

    def _change_ext_conn_vnf_instance(self, vnf_instance_id, request_body,
            http_client=None):
        if http_client is None:
            http_client = self.http_client
        url = os.path.join(
            self.base_vnf_instances_url,
            vnf_instance_id,
            "change_ext_conn")
        resp, body = http_client.do_request(url, "POST",
                body=jsonutils.dumps(request_body))

        return resp, body

    def _rollback_op_occs(self, vnf_lcm_op_occs_id, http_client=None):
        if http_client is None:
            http_client = self.http_client
        rollback_url = os.path.join(
            self.base_vnf_lcm_op_occs_url,
            vnf_lcm_op_occs_id, 'rollback')
        resp, response_body = http_client.do_request(
            rollback_url, "POST")

        return resp, response_body

    def _fail_op_occs(self, vnf_lcm_op_occs_id, http_client=None):
        if http_client is None:
            http_client = self.http_client
        fail_url = os.path.join(
            self.base_vnf_lcm_op_occs_url,
            vnf_lcm_op_occs_id, 'fail')
        resp, response_body = http_client.do_request(
            fail_url, "POST")

        return resp, response_body

    def _retry_op_occs(self, vnf_lcm_op_occs_id, http_client=None):
        if http_client is None:
            http_client = self.http_client
        retry_url = os.path.join(
            self.base_vnf_lcm_op_occs_url,
            vnf_lcm_op_occs_id, 'retry')
        resp, response_body = http_client.do_request(
            retry_url, "POST")

        return resp, response_body

    def _show_op_occs(self, vnf_lcm_op_occs_id, http_client=None):
        if http_client is None:
            http_client = self.http_client
        show_url = os.path.join(
            self.base_vnf_lcm_op_occs_url,
            vnf_lcm_op_occs_id)
        resp, response_body = http_client.do_request(
            show_url, "GET")

        return resp, response_body

    def _wait_terminate_vnf_instance(self, id, timeout=None, http_client=None):
        if http_client is None:
            http_client = self.http_client
        start_time = int(time.time())

        self._wait_vnf_instance(id,
            http_client,
            instantiation_state=fields.VnfInstanceState.NOT_INSTANTIATED,
            timeout=timeout)

        # If gracefulTerminationTimeout is set, check whether vnf
        # instantiation_state is set to NOT_INSTANTIATED after
        # gracefulTerminationTimeout seconds.
        if timeout and int(time.time()) - start_time < timeout:
            self.fail("Vnf is terminated before graceful termination"
                      "timeout period")
        else:
            return

        # wait for status completion
        time.sleep(VNF_DELETE_COMPLETION_WAIT)

    def _get_heat_stack(self, vnf_instance_id, h_client=None,
            prefix_id='vnflcm_'):
        if h_client is None:
            h_client = self.h_client
        try:
            stacks = h_client.stacks.list()
        except Exception:
            return None

        target_stack_name = prefix_id + vnf_instance_id
        target_stakcs = list(
            filter(
                lambda x: x.stack_name == target_stack_name,
                stacks))

        if len(target_stakcs) == 0:
            return None

        return target_stakcs[0]

    def _delete_heat_stack(self, stack_id, h_client=None):
        if h_client is None:
            h_client = self.h_client
        h_client.stacks.delete(stack_id)

    def _wait_until_stack_ready(self, stack_id, expected_status,
            h_client=None):
        if h_client is None:
            h_client = self.h_client
        start_time = time.time()
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)

        while True:
            stack = h_client.stacks.get(stack_id)
            actual_status = stack.stack_status
            print(
                ("Wait:callback_url=<%s>, " +
                "wait_status=<%s> ") %
                (callback_url, actual_status),
                flush=True)

            if actual_status == expected_status:
                return None

            if time.time() - start_time > VNF_LCM_DONE_TIMEOUT:
                if actual_status:
                    error = (
                        "LCM incomplete timeout, " +
                        " stack %(stack_id)s" +
                        " is %(actual)s," +
                        "expected status should be %(expected)s")
                    self.fail(
                        error % {
                            "stack_id": stack_id,
                            "expected": expected_status,
                            "actual": actual_status})
                else:
                    self.fail("LCM incomplete timeout")

            time.sleep(RETRY_WAIT_TIME)

    def _get_heat_resource_list(self, stack_id, nested_depth=0,
            h_client=None):
        if h_client is None:
            h_client = self.h_client
        try:
            resources = h_client.resources.list(
                stack_id, nested_depth=nested_depth)
        except Exception:
            return None

        return resources

    def _get_heat_resource(self, stack_id, resource_name, h_client=None):
        if h_client is None:
            h_client = self.h_client
        try:
            resource = h_client.resources.get(
                stack_id, resource_name)
        except Exception:
            return None

        return resource

    def _get_heat_stack_template(self, stack_id, nested_depth=0,
            h_client=None):
        if h_client is None:
            h_client = self.h_client
        try:
            template = h_client.stacks.template(stack_id)
        except Exception:
            return None

        return template

    def _get_image_id_from_resource_attributes(self, stack_resource_details):
        if stack_resource_details is None:
            return None
        if not hasattr(stack_resource_details, 'attributes'):
            return None

        return stack_resource_details.attributes.get('image', {}).get('id')

    def _get_vnfc_instance_id_list(
            self,
            stack_id,
            resource_type='OS::Nova::Server',
            nested_depth=2,
            limit=2,
            h_client=None):
        if h_client is None:
            h_client = self.h_client
        resources = self._get_heat_resource_list(
            stack_id, nested_depth=nested_depth, h_client=h_client)
        if resources is None:
            return None

        return [r.physical_resource_id for r in resources[:limit]
            if r.resource_type == resource_type]

    def assert_http_header_location_for_create(self, response_header):
        """Validate URI in location header for CreateVNF

        {apiRoot}/vnflcm/v1/vnf_instances/{vnfInstanceId}/instantiate
        """
        location = response_header.get(
            "Location") or response_header.get("location")
        self.assertIsNotNone(location)
        uri = urlparse(location)
        self.assertIn(uri.scheme, ['http', 'https'])
        self.assertRegex(
            uri.path,
            r'^/(?P<apiRoot>[^/]*?/)?vnflcm/v1/vnf_instances/' +
            UUID_RE)

    def assert_http_header_location_for_lcm_op_occs(self, response_header):
        """Validate URI in location header for various LCMs

        {apiRoot}/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}
        """
        location = response_header.get(
            "Location") or response_header.get("location")
        self.assertIsNotNone(location)
        uri = urlparse(location)
        self.assertIn(uri.scheme, ['http', 'https'])
        self.assertRegex(
            uri.path,
            r'^/(?P<apiRoot>[^/]*?/)?vnflcm/v1/vnf_lcm_op_occs/' +
            UUID_RE)

    def assert_http_header_location_for_subscription(self, response_header):
        """Validate URI in location header for Subscription

        {apiRoot}/vnflcm/v1/subscriptions/{subscriptionId}
        """
        location = response_header.get(
            "Location") or response_header.get("location")
        self.assertIsNotNone(location)
        uri = urlparse(location)
        self.assertIn(uri.scheme, ['http', 'https'])
        self.assertRegex(
            uri.path,
            r'^/(?P<apiRoot>[^/]*?/)?vnflcm/v1/subscriptions/' +
            UUID_RE)

    def assert_instantiation_state(
            self,
            vnf_instance_body,
            expected_instantiation_state=fields.VnfInstanceState.INSTANTIATED):
        # FT-checkpoint: Instantiation state(VNF instance)
        self.assertEqual(
            expected_instantiation_state,
            vnf_instance_body['instantiationState'])

    def assert_vnf_state(
            self,
            vnf_instance_body,
            expected_vnf_state=fields.VnfOperationalStateType.STARTED):
        # FT-checkpoint: vnf_state
        self.assertEqual(
            expected_vnf_state,
            vnf_instance_body['instantiatedVnfInfo']['vnfState'])

    def assert_heat_stack_status(
            self,
            vnf_instance_id,
            h_client=None,
            expected_stack_status=infra_cnst.STACK_CREATE_COMPLETE):
        if h_client is None:
            h_client = self.h_client
        stack = self._get_heat_stack(vnf_instance_id, h_client)
        self.assertEqual(
            expected_stack_status,
            stack.stack_status)

    def assert_heat_resource_status(
            self,
            vnf_instance,
            h_client=None,
            expected_glance_image=None,
            expected_resource_status=None):

        if h_client is None:
            h_client = self.h_client

        def assert_glance_image(stack_id, resource_name, h_client):
            resource_details = self._get_heat_resource(stack_id,
                    resource_name,
                    h_client)
            image = self._get_image_id_from_resource_attributes(
                resource_details)
            if image:
                self.assertEqual(expected_glance_image, image.status)

        stack = self._get_heat_stack(vnf_instance['id'], h_client)
        resources = self._get_heat_resource_list(stack.id,
                nested_depth=2, h_client=h_client)
        self.assertIsNotNone(resources)

        for resource in resources:
            # FT-checkpoint: resource status
            self.assertEqual(expected_resource_status,
                resource.resource_status)

            # FT-checkpoint: Glance-image
            if expected_glance_image:
                assert_glance_image(stack.id, resource.resource_name,
                        h_client)

    def assert_heat_resource_status_is_none(
            self,
            stack_id,
            h_client=None,
            glance_client=None,
            resources_name_list=None,
            glance_image_id_list=None):
        if h_client is None:
            h_client = self.h_client
        if glance_client is None:
            glance_client = self.glance_client
        resources_name_list = resources_name_list or []
        for resource_name in resources_name_list:
            resource = self._get_heat_resource(stack_id,
                    resource_name,
                    h_client)
            self.assertIsNone(resource)

        glance_image_id_list = glance_image_id_list or []
        for glance_image_id in glance_image_id_list:
            image = self._get_glance_image(glance_image_id,
                    glance_client)
            self.assertIsNone(image)

    def _wait_lcm_done(self,
            expected_operation_status=None,
            vnf_instance_id=None,
            fake_server_manager=None):
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        start_time = int(time.time())
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)

        while True:
            actual_status = None
            vnf_lcm_op_occ_id = None
            notify_mock_responses = fake_server_manager.get_history(
                callback_url)
            print(
                ("Wait:callback_url=<%s>, " +
                "wait_status=<%s>, " +
                "vnf_instance_id=<%s>") %
                (callback_url, expected_operation_status, vnf_instance_id),
                flush=True)

            for res in notify_mock_responses:
                if vnf_instance_id != res.request_body.get('vnfInstanceId'):
                    continue

                if expected_operation_status is None:
                    return

                actual_status = res.request_body.get('operationState', '')
                vnf_lcm_op_occ_id = res.request_body.get('vnfLcmOpOccId', '')
                if actual_status == expected_operation_status:
                    return

            if ((int(time.time()) - start_time) > VNF_LCM_DONE_TIMEOUT):
                if actual_status:
                    error = (
                        "LCM incomplete timeout, %(vnf_lcm_op_occ_id)s" +
                        " is %(actual)s," +
                        "expected status should be %(expected)s")
                    self.fail(
                        error % {
                            "vnf_lcm_op_occ_id": vnf_lcm_op_occ_id,
                            "expected": expected_operation_status,
                            "actual": actual_status})
                else:
                    self.fail("LCM incomplete timeout")

            time.sleep(RETRY_WAIT_TIME)

    def _wait_stack_update(self, vnf_instance_id, expected_status,
            h_client=None):
        if h_client is None:
            h_client = self.h_client
        timeout = VNF_HEAL_TIMEOUT
        start_time = int(time.time())
        while True:
            stack = self._get_heat_stack(vnf_instance_id, h_client)
            if stack.stack_status == expected_status:
                break

            if ((int(time.time()) - start_time) > timeout):
                error = ("Stack %(id)s status is %(current)s, expected status "
                        "should be %(expected)s")
                self.fail(error % {"vnf_instance_name": vnf_instance_id,
                    "current": stack.status,
                    "expected": expected_status})

            time.sleep(RETRY_WAIT_TIME)

    def assert_create_vnf(self, resp, vnf_instance,
            fake_server_manager=None):
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        self.assertEqual(201, resp.status_code)

        self.assert_http_header_location_for_create(resp.headers)
        self.assert_instantiation_state(
            vnf_instance,
            fields.VnfInstanceState.NOT_INSTANTIATED)

        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance.get('id'),
            fake_server_manager=fake_server_manager)

        self.assertEqual(1, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfIdentifierCreationNotification')

    def assert_delete_vnf(self, resp, vnf_instance_id, http_client=None,
            fake_server_manager=None):
        self.assertEqual(204, resp.status_code)

        if http_client is None:
            http_client = self.http_client
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER

        resp, _ = self._show_vnf_instance(vnf_instance_id, http_client)
        self.assertEqual(404, resp.status_code)

        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(1, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfIdentifierDeletionNotification')

    def assert_instantiate_vnf(
            self,
            resp,
            vnf_instance_id,
            http_client=None,
            h_client=None,
            fake_server_manager=None):
        if http_client is None:
            http_client = self.http_client
        if h_client is None:
            h_client = self.h_client
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER

        self.assertEqual(202, resp.status_code)
        resp, vnf_instance = self._show_vnf_instance(vnf_instance_id,
                http_client)
        self.assert_vnf_state(vnf_instance)

        self.assert_heat_stack_status(vnf_instance['id'], h_client)
        self.assert_heat_resource_status(
            vnf_instance,
            h_client,
            expected_glance_image='active',
            expected_resource_status='CREATE_COMPLETE')

        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(3, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfLcmOperationOccurrenceNotification',
            'STARTING')

        self.assert_notification_mock_response(
            notify_mock_responses[1],
            'VnfLcmOperationOccurrenceNotification',
            'PROCESSING')

        self.assert_notification_mock_response(
            notify_mock_responses[2],
            'VnfLcmOperationOccurrenceNotification',
            'COMPLETED')

    def assert_heal_vnf(
            self,
            resp,
            vnf_instance_id,
            expected_stack_status='UPDATE_COMPLETE',
            http_client=None,
            h_client=None,
            fake_server_manager=None):
        if http_client is None:
            http_client = self.http_client
        if h_client is None:
            h_client = self.h_client
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER

        self.assertEqual(202, resp.status_code)

        resp, vnf_instance = self._show_vnf_instance(vnf_instance_id,
                http_client)
        self.assert_vnf_state(vnf_instance)
        self.assert_instantiation_state(vnf_instance)

        self.assert_heat_stack_status(
            vnf_instance['id'],
            h_client,
            expected_stack_status=expected_stack_status)

        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(3, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfLcmOperationOccurrenceNotification',
            'STARTING')

        self.assert_notification_mock_response(
            notify_mock_responses[1],
            'VnfLcmOperationOccurrenceNotification',
            'PROCESSING')

        self.assert_notification_mock_response(
            notify_mock_responses[2],
            'VnfLcmOperationOccurrenceNotification',
            'COMPLETED')

    def assert_terminate_vnf(
            self,
            resp,
            vnf_instance_id,
            stack_id,
            resource_name_list,
            glance_image_id_list,
            http_client=None,
            h_client=None,
            glance_client=None,
            fake_server_manager=None):
        if http_client is None:
            http_client = self.http_client
        if h_client is None:
            h_client = self.h_client
        if glance_client is None:
            glance_client = self.glance_client
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER

        self.assertEqual(202, resp.status_code)

        resp, vnf_instance = self._show_vnf_instance(vnf_instance_id,
                http_client)
        self.assert_instantiation_state(
            vnf_instance,
            fields.VnfInstanceState.NOT_INSTANTIATED)

        # FT-checkpoint: Heat stack status.
        stack = self._get_heat_stack(vnf_instance_id, h_client)
        self.assertIsNone(stack)

        self.assert_heat_resource_status_is_none(
            stack_id,
            h_client,
            glance_client,
            resources_name_list=resource_name_list,
            glance_image_id_list=glance_image_id_list)

        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(3, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfLcmOperationOccurrenceNotification',
            'STARTING')

        self.assert_notification_mock_response(
            notify_mock_responses[1],
            'VnfLcmOperationOccurrenceNotification',
            'PROCESSING')

        self.assert_notification_mock_response(
            notify_mock_responses[2],
            'VnfLcmOperationOccurrenceNotification',
            'COMPLETED')

    def assert_scale_vnf(
            self,
            resp,
            vnf_instance_id,
            pre_stack_resource_list,
            post_stack_resource_list,
            scale_type='SCALE_OUT',
            expected_stack_status='CREATE_COMPLETE',
            http_client=None,
            h_client=None,
            fake_server_manager=None):
        if http_client is None:
            http_client = self.http_client
        if h_client is None:
            h_client = self.h_client
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        self.assertEqual(202, resp.status_code)
        self.assert_http_header_location_for_lcm_op_occs(resp.headers)

        resp, vnf_instance = self._show_vnf_instance(vnf_instance_id,
                http_client)
        self.assert_vnf_state(vnf_instance)
        self.assert_instantiation_state(vnf_instance)

        # check: scaling stack resource count
        if scale_type == 'SCALE_OUT':
            self.assertTrue(len(pre_stack_resource_list) <
                            len(post_stack_resource_list))
        else:
            self.assertTrue(len(pre_stack_resource_list) >
                            len(post_stack_resource_list))

        # check scaleStatus
        scale_status = vnf_instance['instantiatedVnfInfo']['scaleStatus']
        self.assertTrue(len(scale_status) > 0)
        for status in scale_status:
            self.assertIsNotNone(status.get('aspectId'))
            self.assertIsNotNone(status.get('scaleLevel'))

        self.assert_heat_stack_status(
            vnf_instance['id'],
            h_client,
            expected_stack_status=expected_stack_status)

        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(3, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfLcmOperationOccurrenceNotification',
            'STARTING')

        self.assert_notification_mock_response(
            notify_mock_responses[1],
            'VnfLcmOperationOccurrenceNotification',
            'PROCESSING')

        self.assert_notification_mock_response(
            notify_mock_responses[2],
            'VnfLcmOperationOccurrenceNotification',
            'COMPLETED')

    def assert_rollback_vnf(self, resp, vnf_instance_id,
            fake_server_manager=None):
        self.assertEqual(202, resp.status_code)

        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(2, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfLcmOperationOccurrenceNotification',
            'ROLLING_BACK')

        self.assert_notification_mock_response(
            notify_mock_responses[1],
            'VnfLcmOperationOccurrenceNotification',
            'ROLLED_BACK')

    def assert_fail_vnf(self, resp, vnf_instance_id,
            fake_server_manager=None):
        self.assertEqual(200, resp.status_code)

        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(1, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfLcmOperationOccurrenceNotification',
            'FAILED')

    def assert_retry_vnf(self, resp, vnf_instance_id,
            fake_server_manager=None):
        self.assertEqual(202, resp.status_code)

        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(2, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfLcmOperationOccurrenceNotification',
            'PROCESSING')

        self.assert_notification_mock_response(
            notify_mock_responses[1],
            'VnfLcmOperationOccurrenceNotification',
            'FAILED_TEMP')

    def assert_update_vnf(
            self,
            resp,
            vnf_instance_id,
            expected_stack_status='CREATE_COMPLETE',
            http_client=None,
            h_client=None,
            fake_server_manager=None):
        if http_client is None:
            http_client = self.http_client
        if h_client is None:
            h_client = self.h_client
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        self.assertEqual(202, resp.status_code)
        self.assert_http_header_location_for_lcm_op_occs(resp.headers)

        resp, vnf_instance = self._show_vnf_instance(vnf_instance_id,
                http_client)
        self.assertEqual(200, resp.status_code)

        self.assert_vnf_state(vnf_instance)
        self.assert_instantiation_state(vnf_instance)

        self.assert_heat_stack_status(
            vnf_instance['id'],
            h_client,
            expected_stack_status=expected_stack_status)

        # FT-checkpoint: Notification
        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = self._filter_notify_history(callback_url,
            vnf_instance_id, fake_server_manager=fake_server_manager)

        self.assertEqual(2, len(notify_mock_responses))
        self.assert_notification_mock_response(
            notify_mock_responses[0],
            'VnfLcmOperationOccurrenceNotification',
            'PROCESSING')

        self.assert_notification_mock_response(
            notify_mock_responses[1],
            'VnfLcmOperationOccurrenceNotification',
            'COMPLETED')

    def assert_notification_get(self, callback_url,
            fake_server_manager=None):
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        notify_mock_responses = fake_server_manager.get_history(
            callback_url)
        fake_server_manager.clear_history(
            callback_url)
        self.assertEqual(1, len(notify_mock_responses))
        self.assertEqual(204, notify_mock_responses[0].status_code)

    def assert_notification_mock_response(
            self,
            notify_mock_response,
            expected_notify_types,
            expected_operation_status=None):
        self.assertEqual(204, notify_mock_response.status_code)

        self.assertEqual(
            expected_notify_types,
            notify_mock_response.request_body['notificationType'])

        if expected_operation_status:
            self.assertEqual(
                expected_operation_status,
                notify_mock_response.request_body['operationState'])

    def _create_network(self, name, neutron_client=None):
        # First, we have to check network name passed by caller is
        # already exists or not.
        # OK, we can create this.

        if neutron_client is None:
            neutron_client = self.neutronclient()
        try:
            uniq_name = name + '-' + uuidutils.generate_uuid()
            net = \
                self.neutronclient().create_network(
                    {'network': {'name': uniq_name}})
            net_id = net['network']['id']
            self.addCleanup(self._delete_network, net_id)
            print("Create network success, %s" % uniq_name, flush=True)
            return net_id, uniq_name
        except Exception as e:
            self.fail("Failed, create network=<%s>, %s" %
                (uniq_name, e))

    def _create_subnet(self, network, cidr, gateway, neutron_client=None):
        if neutron_client is None:
            neutron_client = self.neutronclient()
        body = {'subnet': {'network_id': network['id'],
                'name': "subnet-%s" % uuidutils.generate_uuid(),
                'cidr': "{}".format(cidr),
                'ip_version': 4,
                'gateway_ip': "{}".format(gateway),
                "enable_dhcp": True}}

        try:
            subnet = neutron_client.create_subnet(body=body)["subnet"]
            self.addCleanup(self._delete_subnet, subnet['id'], neutron_client)
            print("Create subnet success, %s" % subnet['id'], flush=True)
            return subnet['id']
        except Exception as e:
            self.fail("Failed, create subnet for net_id=<%s>, %s" %
                (network['id'], e))

    def _create_port(self, network_id, neutron_client=None):
        if neutron_client is None:
            neutron_client = self.neutronclient()
        body = {'port': {'network_id': network_id}}
        try:
            port = neutron_client.create_port(body=body)["port"]
            self.addCleanup(self._delete_port, port['id'], neutron_client)
            return port['id']
        except Exception as e:
            self.fail("Failed, create port for net_id=<%s>, %s" %
                (network_id, e))

    def _delete_network(self, network_id, neutron_client=None):
        if neutron_client is None:
            neutron_client = self.neutronclient()
        try:
            neutron_client.delete_network(network_id)
        except Exception:
            print("Failed, delete network.", network_id, flush=True)

    def _delete_subnet(self, subnet_id, neutron_client=None):
        if neutron_client is None:
            neutron_client = self.neutronclient()
        try:
            neutron_client.delete_subnet(subnet_id)
        except Exception:
            print("Failed, delete subnet.", subnet_id, flush=True)

    def _delete_port(self, port_id, neutron_client):
        if neutron_client is None:
            neutron_client = self.neutronclient()
        try:
            neutron_client.delete_port(port_id)
        except Exception:
            print("Failed, delete port.", port_id, flush=True)

    def assert_subscription_show(self, resp, response_body):
        """Assert that subscription informations has mandatory keys."""
        self.assertEqual(200, resp.status_code)

        self.assertIsNotNone(response_body.get('id'))
        _filter = response_body.get('filter')
        self.assertIsNotNone(_filter)
        self.assertIsNotNone(_filter.get('notificationTypes'))
        self.assertIsNotNone(_filter.get('operationTypes'))
        self.assertIsNotNone(response_body.get('callbackUri'))
        _links = response_body.get('_links')
        self.assertIsNotNone(_links)
        self.assertIsNotNone(_links.get('self'))
        self.assertIsNotNone(_links.get('self').get('href'))

    def _filter_notify_history(self, callback_url, vnf_instance_id,
            fake_server_manager=None):
        if fake_server_manager is None:
            fake_server_manager = FAKE_SERVER_MANAGER
        notify_histories = fake_server_manager.get_history(
            callback_url)
        fake_server_manager.clear_history(callback_url)

        return [
            h for h in notify_histories
            if h.request_body.get('vnfInstanceId') == vnf_instance_id]

    def _get_heat_stack_show(self, vnf_instance_id, resource_name=None,
            h_client=None):
        """Retrieve image name of the resource from stack"""
        if h_client is None:
            h_client = self.h_client
        try:
            stack = self._get_heat_stack(vnf_instance_id, h_client)
            stack_info = h_client.stacks.get(stack.id)
            stack_dict = stack_info.to_dict()
            resource_dict = json.loads(stack_dict['parameters']['nfv'])
        except Exception:
            return None

        if resource_name is None:
            return resource_dict

        return resource_dict['VDU'][resource_name]['image']



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Request Body validating middleware.

"""

import functools
import webob

from tacker.api.validation import validators
from tacker.common import exceptions


def schema(request_body_schema):
    """Register a schema to validate request body.

    Registered schema will be used for validating request body just before
    API method executing.

    :param dict request_body_schema: a schema to validate request body

    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            schema_validator = validators._SchemaValidator(
                request_body_schema)
            try:
                schema_validator.validate(kwargs['body'])
            except KeyError:
                raise webob.exc.HTTPBadRequest(
                    explanation=_("Malformed request body"))

            return func(*args, **kwargs)
        return wrapper

    return add_validator


def query_schema(query_params_schema):
    """Register a schema to validate request query parameters.

    Registered schema will be used for validating request query params just
    before API method executing.

    :param query_params_schema: A dict, the JSON-Schema for validating the
                                query parameters.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # NOTE(tpatil): The second argument of the method
            # calling this method should always be 'request'.
            if 'request' in kwargs:
                req = kwargs['request']
            else:
                req = args[1]

            try:
                req.GET.dict_of_lists()
            except UnicodeDecodeError:
                msg = _('Query string is not UTF-8 encoded')
                raise exceptions.ValidationError(msg)

            query_opts = {}
            query_opts.update(req.GET)
            schema_validator = validators._SchemaValidator(
                query_params_schema)
            schema_validator.validate(query_opts)

            return func(*args, **kwargs)

        return wrapper

    return add_validator



# Copyright (c) 2012 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from tacker.common import utils
from tacker.tests import base


class TestDict2Tuples(base.BaseTestCase):
    def test_dict(self):
        input_dict = {'foo': 'bar', '42': 'baz', 'aaa': 'zzz'}
        expected = (('42', 'baz'), ('aaa', 'zzz'), ('foo', 'bar'))
        output_tuple = utils.dict2tuple(input_dict)
        self.assertEqual(expected, output_tuple)


class TestChangeMemory(testtools.TestCase):
    def test_change_memory_from_mb_to_gb(self):
        actual_val = utils.change_memory_unit("1024 mb", "GB")
        expected_val = 1
        self.assertEqual(expected_val, actual_val)

    def test_change_memory_from_gb_to_mb(self):
        actual_val = utils.change_memory_unit("1 GB", "MB")
        expected_val = 1024
        self.assertEqual(expected_val, actual_val)


class TestCamelToSnakeCase(testtools.TestCase):
    def test_convert_camelcase_to_snakecase_dict(self):
        """Only the dict keys should be converted to snakecase"""
        actual_val = utils.convert_camelcase_to_snakecase(
            {"camelCaseKey": "camelCaseValue"})
        expected_val = {"camel_case_key": "camelCaseValue"}
        self.assertEqual(expected_val, actual_val)

    def test_convert_camelcase_to_snakecase_list_with_dict_items(self):
        """Only the dict keys from list should be converted to snakecase"""
        data = [{"camelCaseKey": "camelCaseValue"}]
        actual_val = utils.convert_camelcase_to_snakecase(data)
        expected_val = [{"camel_case_key": "camelCaseValue"}]
        self.assertEqual(expected_val, actual_val)

    def test_convert_camelcase_to_snakecase_list_with_string_items(self):
        """Conversion of camelcase to snakecase should be ignored.

        For simple list with string items, the elements which are actual
        values should be ignored during conversion
        """
        data = ["camelCaseValue1", "camelCaseValue2"]
        actual_val = utils.convert_snakecase_to_camelcase(data)
        expected_val = ["camelCaseValue1", "camelCaseValue2"]
        self.assertEqual(expected_val, actual_val)


class TestSnakeToCamelCase(testtools.TestCase):
    def test_convert_snakecase_to_camelcase_dict(self):
        """Only the dict keys from list should be converted to camelcase"""
        data = {"snake_case_key": "snake_case_value",
                "_key": "ignore_key_value"}
        actual_val = utils.convert_snakecase_to_camelcase(data)
        expected_val = {"snakeCaseKey": "snake_case_value",
                        "_key": "ignore_key_value"}
        self.assertEqual(expected_val, actual_val)

    def test_convert_snakecase_to_camelcase_list_with_dict_items(self):
        """Only the dict keys from list should be converted to camelcase"""
        data = [{"snake_case_key": "snake_case_value"}]
        actual_val = utils.convert_snakecase_to_camelcase(data)
        expected_val = [{"snakeCaseKey": "snake_case_value"}]
        self.assertEqual(expected_val, actual_val)

    def test_convert_snakecase_to_camelcase_list_with_string_items(self):
        """Conversion of snakecase to camelcase should be ignored.

        For simple list with string items, the elements which are actual
        values should be ignored during conversion
        """
        data = ["snake_case_value1", "snake_case_value2"]
        actual_val = utils.convert_snakecase_to_camelcase(data)
        expected_val = ["snake_case_value1", "snake_case_value2"]
        self.assertEqual(expected_val, actual_val)


class TestValidateUrl(testtools.TestCase):
    def test_valid_url(self):
        result = utils.is_valid_url("https://10.10.10.10/test.zip")
        self.assertTrue(result)

    def test_no_scheme(self):
        result = utils.is_valid_url("//10.10.10.10/test.zip")
        self.assertFalse(result)

    def test_invalid_scheme(self):
        result = utils.is_valid_url("invalid://10.10.10.10/test.zip")
        self.assertFalse(result)

    def test_no_path_in_url(self):
        # The specified url `https://10.10.10.10` is valid but in context with
        # the functionality implemented, expecting csar file in `path` as
        # mandatory parameter.
        result = utils.is_valid_url("https://10.10.10.10")
        self.assertFalse(result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from tacker._i18n import _
from tacker.common import exceptions
from tacker.services import service_base


class NSPluginBase(service_base.NFVPluginBase, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create_nsd(self, context, nsd):
        pass

    @abc.abstractmethod
    def delete_nsd(self, context, nsd_id):
        pass

    @abc.abstractmethod
    def get_nsd(self, context, nsd_id, fields=None):
        pass

    @abc.abstractmethod
    def get_nsds(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def create_ns(self, context, ns):
        pass

    @abc.abstractmethod
    def get_nss(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_ns(self, context, ns_id, fields=None):
        pass

    @abc.abstractmethod
    def delete_ns(self, context, ns_id):
        pass


class NSDNotFound(exceptions.NotFound):
    message = _('NSD %(nsd_id)s could not be found')


class NSNotFound(exceptions.NotFound):
    message = _('NS %(ns_id)s could not be found')


class NSInUse(exceptions.InUse):
    message = _('NS %(ns_id)s in use')



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

from oslo_serialization import jsonutils

from tacker.sol_refactored.objects import base
from tacker.sol_refactored.objects import fields
from tacker.tests import base as tests_base


@base.TackerObjectRegistry.register
class MyObj(base.TackerObject):

    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'data': fields.StringField(nullable=False),
        'listData': fields.ListOfObjectsField(
            'MySubObj', nullable=True),
        'createdAt': fields.DateTimeField(nullable=False),
    }


@base.TackerObjectRegistry.register
class MySubObj(base.TackerObject):

    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'data': fields.StringField(nullable=False),
    }


@base.TackerObjectRegistry.register
class MyDBObj(MyObj, base.TackerPersistentObject):

    pass


class TestTackerObject(tests_base.BaseTestCase):

    def test_tacker_obj_get_changes(self):
        o = MyObj.from_dict({'id': 'foo',
                             'data': 'abcde',
                             'createdAt': '2021-09-01T12:34:56+09:00'})
        o.obj_reset_changes()
        self.assertEqual(o.tacker_obj_get_changes(), {})
        o.data = '12345'
        o.createdAt = datetime.datetime(2021, 8, 7, 6, 5, 44,
            tzinfo=datetime.timezone(datetime.timedelta(hours=-9)))
        changes = o.tacker_obj_get_changes()
        self.assertEqual(len(changes), 2)
        self.assertIn('data', changes)
        self.assertIsNone(changes['createdAt'].tzinfo)
        self.assertEqual(changes['createdAt'].hour, o.createdAt.hour + 9)

    def test_from_dict(self):
        o = MyObj.from_dict({'id': 'foo',
                             'data': 'abcde',
                             'listData': [
                                 {'id': 'foo1', 'data': 'bar1'},
                                 {'id': 'foo2', 'data': 'bar2'},
                             ],
                             'createdAt': '2021-09-01T12:34:56+09:00'})
        self.assertEqual(o.id, 'foo')
        self.assertIsInstance(o.createdAt, datetime.datetime)
        self.assertEqual(len(o.listData), 2)
        self.assertEqual(o.listData[1].data, 'bar2')


class TestTackerObjectSerializer(tests_base.BaseTestCase):

    def test_serialize_entity(self):
        serializer = base.TackerObjectSerializer()
        o = MyObj.from_dict({'id': 'foo',
                             'data': 'abcde',
                             'listData': [
                                 {'id': 'foo1', 'data': 'bar1'},
                                 {'id': 'foo2', 'data': 'bar2'},
                             ],
                             'createdAt': '2021-09-01T12:34:56+09:00'})
        entity = serializer.serialize_entity(mock.Mock(), o)
        self.assertEqual(entity['tacker_sol_refactored_object.name'], 'MyObj')
        self.assertEqual(entity['tacker_sol_refactored_object.namespace'],
            'tacker_sol_refactored')
        data = entity['tacker_sol_refactored_object.data']
        self.assertEqual(set(data.keys()),
                         set(['id', 'data', 'listData', 'createdAt']))

        o2 = serializer.deserialize_entity(mock.Mock(), entity)
        self.assertEqual(o2.listData[1].id, o.listData[1].id)
        self.assertEqual(o2.createdAt, o.createdAt)


class TestTackerPersistentObject(tests_base.BaseTestCase):

    def test_from_db_obj(self):
        o = MyDBObj.from_db_obj(
            {'id': 'foo', 'data': 'abcde',
             'listData': '[{"id": "foo1", "data": "bar1"},'
             '{"id": "foo2", "data": "bar2"}]',
             'createdAt': datetime.datetime(2021, 9, 1, 12, 34, 56)})
        self.assertEqual(o.id, 'foo')
        self.assertEqual(len(o.listData), 2)
        self.assertEqual(o.listData[0].data, 'bar1')

    def test_to_db_obj(self):
        o = MyDBObj.from_dict({'id': 'foo',
                               'data': 'abcde',
                               'listData': [
                                   {'id': 'foo1', 'data': 'bar1'},
                                   {'id': 'foo2', 'data': 'bar2'},
                               ],
                               'createdAt': '2021-09-01T12:34:56'})
        dbobj = o.to_db_obj()
        self.assertEqual(jsonutils.loads(dbobj['listData']),
            [{"id": "foo1", "data": "bar1"}, {"id": "foo2", "data": "bar2"}])



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker.objects import base
from tacker.objects import fields

LOG = logging.getLogger(__name__)


@base.TackerObjectRegistry.register
class TerminateVnfRequest(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'termination_type': fields.VnfInstanceTerminationTypeField(
            nullable=False),
        'graceful_termination_timeout': fields.IntegerField(nullable=True,
                                                            default=0),
        'additional_params': fields.DictOfNullableField(nullable=True,
            default={}),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_terminate_vnf_req = super(
                TerminateVnfRequest, cls).obj_from_primitive(
                primitive, context)
        else:
            obj_terminate_vnf_req = TerminateVnfRequest._from_dict(primitive)

        return obj_terminate_vnf_req

    @classmethod
    def _from_dict(cls, data_dict):
        termination_type = data_dict.get('termination_type')
        graceful_termination_timeout = \
            data_dict.get('graceful_termination_timeout', 0)
        additional_params = data_dict.get('additional_params', {})

        return cls(termination_type=termination_type,
                   graceful_termination_timeout=graceful_termination_timeout,
                   additional_params=additional_params)



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# flake8: noqa: E402

"""mod_vnflcm_subscription

Revision ID: d25c7c865ce8
Revises: 2c5211036579
Create Date: 2020-10-15 14:27:04.946002

"""

# revision identifiers, used by Alembic.
revision = 'd25c7c865ce8'
down_revision = '2c5211036579'

from alembic import op
import sqlalchemy as sa
from tacker.db import types

from tacker.db import migration

def upgrade(active_plugins=None, options=None):

    op.alter_column('vnf_lcm_filters', 'subscription_uuid',
        type_=types.Uuid(length=36), existing_type=sa.String(length=255),
        nullable=False)

    sta_str = "json_unquote(json_extract('filter','$.operationTypes'))"
    op.add_column(
        'vnf_lcm_filters',
        sa.Column('operation_types',
                  sa.LargeBinary(length=65536),
                  sa.Computed(sta_str)))

    op.add_column(
        'vnf_lcm_filters',
        sa.Column('operation_types_len',
                  sa.Integer,
                  sa.Computed("ifnull(json_length('operation_types'),0)")))

    op.drop_column('vnf_lcm_filters', 'operation_states')
    op.drop_column('vnf_lcm_filters', 'operation_states_len')

    op.alter_column('vnf_lcm_op_occs', 'operation_state',
       type_=sa.String(length=16), existing_type=sa.String(length=255))

    op.alter_column('vnf_lcm_op_occs', 'operation',
        type_=sa.String(length=16),existing_type=sa.String(length=255))

    op.add_column('vnf_lcm_op_occs',
        sa.Column('is_cancel_pending', sa.Boolean, nullable=False)),

    op.add_column('vnf_lcm_op_occs',
        sa.Column('resource_changes', sa.JSON(), nullable=True))

    op.add_column('vnf_lcm_op_occs',
        sa.Column('error_point', sa.Integer, nullable=True))

    op.add_column('vnf_lcm_op_occs',
        sa.Column('changed_info', sa.JSON(), nullable=True))

    op.add_column('vnf_lcm_op_occs',
        sa.Column('created_at', sa.DateTime(), nullable=False))

    op.add_column('vnf_lcm_op_occs',
        sa.Column('updated_at', sa.DateTime(), nullable=True))

    op.add_column('vnf_lcm_op_occs',
        sa.Column('deleted_at', sa.DateTime(), nullable=True))



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import shutil
import subprocess


# vnfdId of samples must be this.
SAMPLE_VNFD_ID = "b1bb0ce7-ebca-4fa7-95ed-4840d7000000"


def make_zip(sample_dir, tmp_dir, vnfd_id, image_path=None):
    # NOTE: '.zip' will be added by shutil.make_archive
    zip_file_name = os.path.basename(os.path.abspath(sample_dir))
    zip_file_path = os.path.join(tmp_dir, zip_file_name)

    tmp_contents = os.path.join(tmp_dir, "contents")
    shutil.copytree(os.path.join(sample_dir, "contents"), tmp_contents)

    # add common vnfd files
    common_dir = os.path.join(sample_dir, "../common/Definitions/")
    for entry in os.listdir(common_dir):
        shutil.copy(os.path.join(common_dir, entry),
                    os.path.join(tmp_contents, "Definitions"))

    # replace vnfd_id
    def_path = os.path.join(tmp_contents, "Definitions")
    for entry in os.listdir(def_path):
        entry_path = os.path.join(def_path, entry)
        with open(entry_path, 'r') as f:
            content = f.read()
        content = content.replace(SAMPLE_VNFD_ID, vnfd_id)
        with open(entry_path, 'w') as f:
            f.write(content)

    if image_path is not None:
        # mkdir Files/ and copy image_path into it
        file_path = os.path.join(tmp_contents, "Files", "images")
        os.makedirs(file_path)
        shutil.copy(image_path, file_path)

    shutil.make_archive(zip_file_path, "zip", tmp_contents)


def create_network(network):
    # assume OS_* environment variables are already set
    subprocess.run(
        ["openstack", "net", "create", network])


def delete_network(network):
    # assume OS_* environment variables are already set
    subprocess.run(
        ["openstack", "net", "delete", network])


def get_network_ids(networks):
    # assume OS_* environment variables are already set
    net_ids = {}
    for net in networks:
        p = subprocess.run(
            ["openstack", "net", "show", net, "-c", "id", "-f", "json"],
            capture_output=True, encoding='utf-8')
        net_ids[net] = json.loads(p.stdout)['id']
    return net_ids


def create_subnet(subnet, network, sub_range, version):
    # assume OS_* environment variables are already set
    subprocess.run(
        ["openstack", "subnet", "create", subnet, "--network", network,
         "--subnet-range", sub_range, "--ip-version", version])


def get_subnet_ids(subnets):
    # assume OS_* environment variables are already set
    subnet_ids = {}
    for subnet in subnets:
        p = subprocess.run(
            ["openstack", "subnet", "show", subnet, "-c", "id", "-f", "json"],
            capture_output=True, encoding='utf-8')
        subnet_ids[subnet] = json.loads(p.stdout)['id']
    return subnet_ids


def create_port(port, network):
    # assume OS_* environment variables are already set
    subprocess.run(
        ["openstack", "port", "create", port, "--network", network])


def delete_port(port):
    # assume OS_* environment variables are already set
    subprocess.run(
        ["openstack", "port", "delete", port])


def get_port_ids(ports):
    # assume OS_* environment variables are already set
    port_ids = {}
    for port in ports:
        p = subprocess.run(
            ["openstack", "port", "show", port, "-c", "id", "-f", "json"],
            capture_output=True, encoding='utf-8')
        port_ids[port] = json.loads(p.stdout)['id']
    return port_ids



# Copyright 2018 NTT DATA
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

from tacker.plugins.common import constants
from tacker.vnfm.infra_drivers.openstack import heat_client as hc
from tacker.vnfm import utils as vnfm_utils
from tacker.vnfm import vim_client


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class Vdu(object):
    def __init__(self, context, vnf_dict, heal_request_data_obj):
        super(Vdu, self).__init__()
        self.context = context
        self.vnf_dict = vnf_dict
        self.heal_request_data_obj = heal_request_data_obj
        self.stack_id = self.heal_request_data_obj.stack_id
        vim_id = self.vnf_dict['vim_id']
        vim_res = vim_client.VimClient().get_vim(context, vim_id)
        placement_attr = vnf_dict.get('placement_attr', {})
        auth_attr = vim_res['vim_auth']
        region_name = placement_attr.get('region_name', None)
        self.heat_client = hc.HeatClient(auth_attr=auth_attr,
                                         region_name=region_name)

    def _get_resource_status(self, stack_id, rsc_name):
        # Get the status of VDU resource from heat
        vdu_resource = self.heat_client.resource_get(stack_id=stack_id,
                                                     rsc_name=rsc_name)
        return vdu_resource.resource_status

    def _resource_mark_unhealthy(self):
        """Mark the resource unhealthy using heat."""

        additional_params = self.heal_request_data_obj.\
            legacy_additional_params
        for additional_param in additional_params:
            resource_name = additional_param.parameter
            res_status = self._get_resource_status(self.stack_id,
                                                   resource_name)
            if res_status != 'CHECK_FAILED':
                self.heat_client.resource_mark_unhealthy(
                    stack_id=self.stack_id,
                    resource_name=resource_name, mark_unhealthy=True,
                    resource_status_reason=additional_param.cause)
                LOG.debug("Heat stack '%s' resource '%s' marked as "
                          "unhealthy", self.stack_id,
                          resource_name)
                evt_details = (("HealVnfRequest invoked to mark resource "
                                "'%s' to unhealthy.") % resource_name)
                vnfm_utils.log_events(self.context, self.vnf_dict,
                                      constants.RES_EVT_HEAL,
                                      evt_details)
            else:
                LOG.debug("Heat stack '%s' resource '%s' already mark "
                          "unhealthy.", self.stack_id,
                          resource_name)

    def heal_vdu(self):
        """Update stack using heat.

         This will re-create the resource which are mark as unhealthy.
         """

        # Mark all the resources as unhealthy
        self._resource_mark_unhealthy()
        self.heat_client.update(stack_id=self.stack_id,
                                existing=True)
        LOG.debug("Heat stack '%s' update initiated to revive "
                  "unhealthy resources.", self.stack_id)
        evt_details = (("HealVnfRequest invoked to update the stack "
                        "'%s'") % self.stack_id)
        vnfm_utils.log_events(self.context, self.vnf_dict,
                              constants.RES_EVT_HEAL, evt_details)



# Copyright 2019 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""modify_unique_constraint_on_vnf_packages_user_data

Revision ID: abbef484b34c
Revises: 9d425296f2c3
Create Date: 2019-11-18 19:34:26.853715

"""

# revision identifiers, used by Alembic.
revision = 'abbef484b34c'
down_revision = '9d425296f2c3'

from alembic import op  # noqa: E402


def upgrade(active_plugins=None, options=None):
    op.drop_constraint(
        constraint_name='uniq_vnf_packages_user_data0idid0key0deleted',
        table_name='vnf_packages_user_data',
        type_='unique')

    op.create_unique_constraint(
        constraint_name='uniq_vnf_packages_user_data0package_uuid0key0deleted',
        table_name='vnf_packages_user_data',
        columns=['package_uuid', 'key', 'deleted'])
